package mazegame;

import javax.microedition.lcdui.*;

// * = Wall
//   = Nothing
//  = Ball ($$$)
// S = Start
// F = Finish
// T = Transport taking
// A = Transport appearing
// K = Key
// D = Door
// 1..9 = Monster number 1..9 path start (first occurence of a given number)
// 1..9 = Monster number 1..9 path end (second occurence of a given number)

public class MazegameCanvas 
		extends com.nokia.mid.ui.FullCanvas
		implements Runnable {
	public static final int UP_PRESSED=0x0002;
	public static final int DOWN_PRESSED=0x0040;
	public static final int LEFT_PRESSED=0x0004;
	public static final int RIGHT_PRESSED=0x0020;
	public static final int FIRE_PRESSED=0x0100;
	
	private static final int MILLIS_PER_TICK = 200;
	private Mazegame owner;
	private Thread animationThread;
	private GameStatus status;
	private Image images[];
	private int keyStatus=0;
	private int keyStatusLatch=0;
	private int mapIndex=0;
	private String[] maps={

			"****************" +
			"*S     *   2*" +
			"* *****D** *" +
			"* *7  7*  * *" +
			"* * 8***3   3* *" +
			"* **T** *" +
			"* **4     4* *" +
			"* ** ******** *" +
			"* *    *A5* *" +
			"* ** * *" +
			"* ** * *" +
			"* **6   6* *" +
			"* *K8***5F* *" +
			"* **********2*" +
			"*1           1 *" +
			"****************",
			
			"****************" +
			"*S**1*" +
			"******" +
			"**1**" +
			"***************" +
			"*2            2*" +
			"******" +
			"**************" +
			"**3           3*" +
			"****** *******" +
			"*4       4**" +
			"******* ****D**" +
			"*K*A F***T   **" +
			"* ***** *******" +
			"*5      5*" +
			"****************",
			
			"****************" +
			"*S  *1 *   K*" +
			"*   * 2*    *" +
			"*   *  *    *" +
			"*              *" +
			"**1 *   *" +
			"** 2*   *" +
			"*****D**********" +
			"* *A *" +
			"* *  *" +
			"*     *5    5  *" +
			"* 4  4* 6    6 *" +
			"*3  3 *  7    7*" +
			"*** ******* ****" +
			"*T*   F*" +
			"****************",

			"****************" +
			"*S*F*" +
			"*152*8*" +
			"** *" +
			"*K* *" +
			"** *" +
			"*21* *" +
			"********D***** *" +
			"*34* *" +
			"** *" +
			"** *" +
			"*43  *" +
			"*76* *" +
			"** *" +
			"*657*8*" +
			"****************"				
			};
		
	public MazegameCanvas(Mazegame owner) throws java.io.IOException {		
		super();
		this.owner=owner;		
		this.animationThread=new Thread(this);
		this.status=new GameStatus(maps[mapIndex],0);

		images=new Image[9];
		images[0]=Image.createImage("/bicho.png");
		images[1]=Image.createImage("/bola.png");
		images[2]=Image.createImage("/heroe.png");
		images[3]=Image.createImage("/llave.png");
		images[4]=Image.createImage("/muro.png");
		images[5]=Image.createImage("/puerta.png");
		images[6]=Image.createImage("/salida.png");
		images[7]=Image.createImage("/suelo.png");
		images[8]=Image.createImage("/transporte.png");

		this.animationThread.start();
	}
		
	// Capture key events
	public int getGameAction(int keyCode) {
		switch (keyCode) {
		case KEY_NUM2: case -1: return UP_PRESSED;
		case KEY_NUM8: case -2: return DOWN_PRESSED;
		case KEY_NUM4: case -3: return LEFT_PRESSED;
		case KEY_NUM6: case -4: return RIGHT_PRESSED;
		case KEY_NUM5: case -5: return FIRE_PRESSED;		
		}
		return 0;
	}
	
	protected synchronized void keyPressed(int keyCode) {
		keyStatus|=getGameAction(keyCode);
	}
	
	protected synchronized void keyRepeated(int keyCode) {
		keyStatus|=getGameAction(keyCode);
	}
	
	protected synchronized void keyReleased(int keyCode) {
		keyStatusLatch|=getGameAction(keyCode);
	}
	
	// Offer captured key events, latched
	protected synchronized int getKeyStates() {
		int result=keyStatus;
		keyStatus&=(~keyStatusLatch);
		keyStatusLatch=0;
		return result;
	}
	
	public Image whichImage(char c) {
		int i;
		switch (c) {
			case '@': i=0; break;
			case '': i=1; break;
			case '#':
			case 'S': i=2; break;
			case 'K': i=3; break;
			case '*': i=4; break;
			case 'D': i=5; break;
			case 'F': i=6; break;
			case ' ':			
			case 'A': i=7; break;
			case 'T': i=8; break;
			default: i=7; break; 
		}
		return this.images[i];
	}
	
	public int[] searchPiece (char piece) {		
		for (int y=0;y<16;y++)
			for (int x=0;x<16;x++)
				if (status.board[x][y]==piece) {
					int result[]=new int[2];
					result[0]=x;
					result[1]=y;
					return result;
				}
		return null;		
	}
	
	public void paint(Graphics g) {
		if (status.finished) {
			g.setColor(0xFFFFFF);			
			g.drawString("SCORE = "+status.puntos, 64, 64, Graphics.BASELINE|Graphics.HCENTER);
			g.setColor(0x000000);			
			g.drawString("SCORE = "+status.puntos, 65, 65, Graphics.BASELINE|Graphics.HCENTER);
		} else {
			// First time repaint all board
			if (status.dirtyIndex==-1) {
				for (int y=0;y<16;y++)
					for (int x=0;x<16;x++) {
						g.drawImage(whichImage(status.board[x][y]),8*x,8*y,Graphics.TOP|Graphics.LEFT);
					}						
			} else {
			// Repaint only dirty cells
				for (int i=0;i<status.dirtyIndex;i++) {
					int x=status.dirty[i][0];
					int y=status.dirty[i][1];
					g.drawImage(whichImage(status.board[x][y]),
							8*x,8*y,Graphics.TOP|Graphics.LEFT);
				}			
			}
			status.dirtyIndex=0;			
		}
	}

	public void tick() {
		int keyState=this.getKeyStates();
		
		if (!status.finished) {
			int[] where;
			
			switch (status.heroe.pieceUnder) {
			case '':
				status.heroe.pieceUnder=' ';
				status.puntos+=10;
				break;
			case 'T': 
				if ((where=searchPiece('A'))!=null) {
					status.heroe.moveAbsolute(where[0],where[1],status);
					status.puntos+=100;
				}
				break;
			case 'K':				
				if ((where=searchPiece('D'))!=null) {
					status.heroe.pieceUnder=' ';
					status.board[where[0]][where[1]]=' ';
					status.dirty[status.dirtyIndex][0]=where[0];
					status.dirty[status.dirtyIndex][1]=where[1];
					status.dirtyIndex++;
					status.puntos+=100;
				}
				break;
			case 'F':
				status.puntos+=150;
				mapIndex++;
				if (mapIndex==maps.length) {
				 status.finished=true;
				} else {
				 status=new GameStatus(maps[mapIndex],status.puntos);
				 return;
				}
				break;
			case '@':
				status.finished=true;
				break;
			}
						
			// Restore bichos background
			for (int i=(status.bichos.length-1);i>=0;i--) {
				if (status.bichos[i]==null) continue;
				status.bichos[i].restoreBg(status);
			}

			// Restore heroe background
			status.heroe.restoreBg(status);
			
			// Move bichos
			for (int i=0;((status.bichos[i]!=null)&&(i<status.bichos.length));i++) {
				status.bichos[i].moveNext(status);
			}
			
			if ((keyState&LEFT_PRESSED)!=0) {
				status.heroe.move(-1,0,status);
			} else if ((keyState&RIGHT_PRESSED)!=0) {
				status.heroe.move(+1,0,status);
			} else if ((keyState&UP_PRESSED)!=0) {
				status.heroe.move(0,-1,status);
			} else if ((keyState&DOWN_PRESSED)!=0) {
				status.heroe.move(0,+1,status);
			} else {
				status.heroe.paint(status);
			}

		} else {
			if ((keyState&FIRE_PRESSED)!=0) {
				owner.endGame();
				animationThread=null;
			}
		}
	}
	
	public void run() {
		Thread currentThread=Thread.currentThread();
		while (currentThread == this.animationThread)
        {
			try {
				long startTime = System.currentTimeMillis();
				if (this.isShown())
				{
					tick();
					repaint(0, 0, getWidth(), getHeight());
					serviceRepaints();
				}
				long elapsedTime = System.currentTimeMillis() - startTime;
				if (elapsedTime < MILLIS_PER_TICK)
				{
					synchronized(this)
					{
						wait(MILLIS_PER_TICK - elapsedTime);
					}
				}
				else
				{
					Thread.yield();
				}
			} catch (InterruptedException e) {			
			}
        }
	}
		
	private class Heroe {
		public int x,y;
		public char pieceUnder;
		public Heroe(int x, int y) {
			this.x=x;
			this.y=y;
		}
		 		
		public void move(int dx, int dy, GameStatus gs) {
			int nx=x+dx;
			int ny=y+dy;
			if (nx<0 || nx>15 || ny<0 || ny>15 || gs.board[nx][ny]=='*' || gs.board[nx][ny]=='D' ) return;
			gs.dirty[gs.dirtyIndex][0]=x;
			gs.dirty[gs.dirtyIndex][1]=y;
			gs.dirtyIndex++;
			x=nx; y=ny;
			gs.dirty[gs.dirtyIndex][0]=x;
			gs.dirty[gs.dirtyIndex][1]=y;
			gs.dirtyIndex++;
			pieceUnder=gs.board[x][y];
			gs.board[x][y]='#';
		}

		public void moveAbsolute(int nx, int ny, GameStatus gs) {
			if (nx<0 || nx>15 || ny<0 || ny>15 || gs.board[nx][ny]=='*' || gs.board[nx][ny]=='D') return;
			gs.heroe.restoreBg(gs);
			gs.dirty[gs.dirtyIndex][0]=x;
			gs.dirty[gs.dirtyIndex][1]=y;
			gs.dirtyIndex++;
			x=nx; y=ny;
			gs.dirty[gs.dirtyIndex][0]=x;
			gs.dirty[gs.dirtyIndex][1]=y;
			gs.dirtyIndex++;
			pieceUnder=gs.board[x][y];
			gs.board[x][y]='#';
		}

		void restoreBg(GameStatus gs) {
			gs.board[x][y]=pieceUnder;
		}
		
		void paint(GameStatus gs) {
			gs.board[x][y]='#';
		}
	}
	
	private class Bicho {
		public int x,y;
		public int route[][];
		public int routeIndex;
		public char pieceUnder; 
		
		public Bicho(int x, int y) {
			this.x=x;
			this.y=y;
			route=new int[2][2];
			routeIndex=0;
			route[0][0]=x;
			route[0][1]=y;
			pieceUnder=' ';
		}
		
		public void addRoutePoint(int x, int y) {
			route[1][0]=x;
			route[1][1]=y;
		}
		
		void moveNext(GameStatus gs) {
			if (gs.dirtyIndex>=0 && gs.dirtyIndex<32) {
				gs.dirty[gs.dirtyIndex][0]=x;
				gs.dirty[gs.dirtyIndex][1]=y;
				gs.dirtyIndex++;
			}			
			if (x<route[routeIndex][0]) x++;
			else if (x>route[routeIndex][0]) x--;
			if (y<route[routeIndex][1]) y++;
			else if (y>route[routeIndex][1]) y--;
			if (gs.dirtyIndex>=0 && gs.dirtyIndex<32) {
				gs.dirty[gs.dirtyIndex][0]=x;
				gs.dirty[gs.dirtyIndex][1]=y;
				gs.dirtyIndex++;
			}
			pieceUnder=gs.board[x][y];
			gs.board[x][y]='@';
			if ((x==gs.heroe.x) && (y==gs.heroe.y)) gs.heroe.pieceUnder='@';
			if ((x==route[routeIndex][0])&&(y==route[routeIndex][1]))
				routeIndex=(routeIndex+1)%2;				
		}

		void restoreBg(GameStatus gs) {
			gs.board[x][y]=pieceUnder;
		}
	}
	
	private class GameStatus {
		public static final int MAX_BICHOS=10;
		
		public char board[][];
		public int dirty[][];
		public int dirtyIndex;
		public Heroe heroe;
		public Bicho bichos[];
		public int puntos;
		public boolean finished;		
		
		public GameStatus(String boardAsString, int puntos) {
			this.board=new char[16][16];
			this.bichos=new Bicho[MAX_BICHOS];
			for (int i=0;i<10;i++) {
				this.bichos[i]=null;
			}
			for (int y=0;y<16;y++)
				for (int x=0;x<16;x++) {
					char c=boardAsString.charAt(16*y+x);
					if (c=='S') {
						this.heroe=new Heroe(x,y);
						this.board[x][y]='S';
					} else if (c>='1'&&c<='9') {
						int bichoIndex=(c-'1');
						if (this.bichos[bichoIndex]==null) {
							this.bichos[bichoIndex]=new Bicho(x,y);
						} else {
							this.bichos[bichoIndex].addRoutePoint(x,y);
						}
						this.board[x][y]=' ';
					} else {
						this.board[x][y]=c;
					}
				}
			this.dirty=new int[32][2];
			this.dirtyIndex=-1; // Mark first time repaint
			this.puntos=puntos;
			this.finished=false;
		}
	}
	
}
